(* Copyright 2009 Erkki Seppälä <flux@modeemi.fi> *)
(* Refer to LICENSE.ganame for licensing details *)

open ExtList
open ExtString

type map =
  | Identity
  | Lowercase
  | Uppercase
  | SpaceToUnderscore
  | UnderscoreToSpace
  | ExampleMap of int

type condition =
  | ChangeCase
  | ExtSeparator
  | Separator

type operation =
  | Nop
  | MapCopyChar of map
  | Until of (condition * operation)
  | Skip
  | Insert of (int * int * int)
  | InsertWord of int			(* insert (n%num_of_words + 1)th word from the current input *)
  | InsertSeparator of char

type gen_ctx = {
  gc_examples : (string * string) array;
  gc_strings : string array; (* strings available for Insert *)
}

let rec gen_map ~gen_ctx () =
  match Random.int 6 with
    | 0 -> Identity
    | 1 -> Lowercase
    | 2 -> SpaceToUnderscore
    | 3 -> UnderscoreToSpace
    | 4 -> Uppercase
    | _ when Array.length gen_ctx.gc_examples = 0 -> gen_map ~gen_ctx ()
    | _ -> ExampleMap (Random.int (Array.length gen_ctx.gc_examples))

let random_char chars = chars.[Random.int (String.length chars)]

let gen_separator () = random_char "._- "

let gen_condition () =
  match Random.int 3 with
    | 0 -> ChangeCase
    | 1 -> ExtSeparator
    | _ -> Separator

(* random_range returns a numbe between [v0..v1] *)
let random_range v0 v1 =
  Random.int (v1 - v0 + 1) + v0

let gen_insert ~gen_ctx () =
  let n = Random.int (Array.length gen_ctx.gc_strings) in
  let str_len = String.length gen_ctx.gc_strings.(n) in
  let ofs = random_range 0 (str_len - 1) in
  let len = random_range 1 (str_len - ofs) in
    (n, ofs, len)

let mutate_insert ~gen_ctx insert =
  match Random.int 4, insert with
      (* random string *)
    | 0, (_, ofs, len) -> 
	let n = Random.int (Array.length gen_ctx.gc_strings) in
	let str = gen_ctx.gc_strings.(n) in
	let ofs' = min (String.length str - 1) ofs in
	let len' = min (String.length str - ofs') len in
	  (n, ofs', len')
      (* random offset *)
    | 1, (n, _, len) ->
	let str = gen_ctx.gc_strings.(n) in
	let ofs' = random_range 0 (String.length str - 1) in
	let len' = min (String.length str - ofs') len in
	  (n, ofs', len')
    (* random length *)
    | 2, (n, ofs, _) ->
	let str = gen_ctx.gc_strings.(n) in
	let str_len = String.length str in
	let len = random_range 1 (str_len - ofs) in
	  (n, ofs, len)
    (* random all *)
    | _, _ ->
	gen_insert ~gen_ctx ()

let rec gen_until_operation ~gen_ctx () =
  match Random.int 2 with
    | 0 -> MapCopyChar (gen_map ~gen_ctx ())
    | _ -> Skip

let gen_insert_word () =
  (if Random.bool () then -1 else 1) * Random.int 10

let rec gen_operation ~gen_ctx () =
  match Random.int 7 with
    | 0 -> Nop
    | 1 -> MapCopyChar (gen_map ~gen_ctx ())
    | 2 -> Until (gen_condition (), gen_until_operation ~gen_ctx ())
    | 3 when Array.length gen_ctx.gc_strings > 0 -> 
	Insert (gen_insert ~gen_ctx ())
    | 3 -> gen_operation ~gen_ctx ()
    | 4 -> InsertSeparator (gen_separator ())
    | 5 -> InsertWord (gen_insert_word ())
    | _ -> Skip

let rec mutate_operation ~gen_ctx = function
  | Nop | Skip -> gen_operation ~gen_ctx ()
  | MapCopyChar _ when Random.int 100 < 70 -> MapCopyChar (gen_map ~gen_ctx ())
  | MapCopyChar _ -> gen_operation ~gen_ctx ()
  | Until (condition, operation) when Random.int 100 < 70 ->
      Until (gen_condition (), gen_until_operation ~gen_ctx ())
  | Until _ -> gen_operation ~gen_ctx ()
  | Insert insert when Random.int 100 < 70 ->
      Insert (mutate_insert ~gen_ctx insert)
  | Insert _ -> gen_operation ~gen_ctx ()
  | InsertSeparator _ -> InsertSeparator (gen_separator ())
  | InsertWord _ -> InsertWord (gen_insert_word ())

let cost_of_map ~gen_ctx = function
  | Identity -> 0
  | Lowercase | Uppercase -> 1
  | SpaceToUnderscore | UnderscoreToSpace -> 1
  | ExampleMap _ -> 3

let rec cost_of_operation ~gen_ctx = function
  | Nop -> 0
  | MapCopyChar map -> 1 + cost_of_map ~gen_ctx map
  | Until (_, op) -> 1 + cost_of_operation ~gen_ctx op
  | Insert (_, _, len) -> 10 + 5 * len
  | InsertWord _ -> 20
  | Skip -> 4
  | InsertSeparator _ -> 2

let string_of_map ~gen_ctx = function
  | Identity -> "Identity"
  | Lowercase -> "Lowercase"
  | Uppercase -> "Uppercase"
  | SpaceToUnderscore -> "SpaceToUnderscore"
  | UnderscoreToSpace -> "UnderscoreToSpace"
  | ExampleMap n -> "ExampleMap(" ^ string_of_int n ^ ")"

let string_of_condition = function
  | ChangeCase -> "ChangeCase"
  | Separator -> "Separator"
  | ExtSeparator -> "ExtSeparator"

let rec string_of_operation ~gen_ctx = function
  | Nop -> "Nop"
  | MapCopyChar Identity -> "Copy"
  | MapCopyChar map -> "MapCopy(" ^ string_of_map ~gen_ctx map ^ ")"
  | Until (cond, operation) -> "Until(" ^ string_of_condition cond ^ "," ^ string_of_operation ~gen_ctx operation ^ ")"
  | Insert (n, ofs, len) -> 
      "Insert(" ^ string_of_int n ^ "," ^ string_of_int ofs ^ "," ^ string_of_int len ^ "="
      ^ String.sub gen_ctx.gc_strings.(n) ofs len ^ ")"
  | InsertWord n -> "InsertWord(" ^ string_of_int n ^ ")"
  | Skip -> "Skip"
  | InsertSeparator ch -> "InsertSeparator(" ^ String.make 1 ch ^ ")"

let is_lower ch = ch = Char.lowercase ch

let char_in chars ch =
  try ignore (String.index chars ch); true 
  with Not_found -> false

let map ~gen_ctx m ch = 
  match m with
    | Identity -> ch
    | Lowercase -> Char.lowercase ch
    | Uppercase -> Char.uppercase ch
    | SpaceToUnderscore -> 
	if ch = ' ' then '_'
	else ch
    | UnderscoreToSpace -> 
	if ch = '_' then ' '
	else ch
    | ExampleMap n ->
	let (src, dst) = gen_ctx.gc_examples.(n) in
	  try
	    let idx = String.index src ch in
	      if idx < String.length dst then
		dst.[idx]
	      else
		ch
	  with Not_found -> ch

let operate ~gen_ctx source geno = 
  let b = Buffer.create (String.length source) in
  let ofs = ref 0 in
  let changes = ref 0 in
  let unused = ref 0 in
  let words = lazy (Pcre.asplit ~pat:"[_,.-]" source) in
  let rec until cond operation =
    let rec aux () =
      if !ofs < String.length source then
	if not (cond source.[!ofs]) then
	  ( operate operation;
	    aux () )
    in
      if !ofs < String.length source then
	aux ()
      else
	incr unused
  and operate = function
      | Nop -> ()
      | MapCopyChar m ->
	  if !ofs < String.length source then
	    ( let ch = source.[!ofs] in
	      let ch' = map ~gen_ctx m ch in
		if ch <> ch' then incr changes
		else if m <> Identity then incr unused;
		Buffer.add_char b ch';
		incr ofs )
	  else
	    incr unused
      | Until (cond, operation) ->
	  ( match cond with
	      | ChangeCase -> 
		  if !ofs < String.length source then
		    let ch0_case = is_lower source.[!ofs] in
		      until (fun cur -> is_lower cur <> ch0_case) operation
		  else
		    incr unused
	      | ExtSeparator ->
		  until (char_in ".") operation
	      | Separator ->
		  until (char_in " _-,") operation 
	  )
      | Skip ->
	  if !ofs < String.length source then
	    incr ofs
	  else
	    incr unused
      | Insert (n, ofs, len) ->
	  Buffer.add_substring b gen_ctx.gc_strings.(n) ofs len
      | InsertWord n when n >= 0 ->
	  let words = Lazy.force words in
	  let word = words.(n mod Array.length words) in
	    Buffer.add_string b word;
      | InsertWord n (* when n < 0 *) ->
	  let n = -n in
	  let words = Lazy.force words in
	  let word = words.(Array.length words - 1 - (n - 1) mod Array.length words) in
	    Buffer.add_string b word;
      | InsertSeparator ch ->
	  Buffer.add_char b ch
  in
    for c = 0 to Array.length geno - 1 do
      operate geno.(c)
    done;
    (Buffer.contents b, (!changes, !unused))

let evaluate_genome ~gen_ctx strings geno =
  List.fold_left
    (fun (errors, bloat, best_bloat) (s1, s2) ->
       let str, (changes, unused) = operate ~gen_ctx s1 geno in
       let cost = Array.fold_left (fun sum op -> sum +. float (cost_of_operation ~gen_ctx op) ** 2.0) 0.0 geno in
       let distance = float (FastLevenstein.leven str s2) in
	 (errors +. distance ** 2.0,
	  bloat +. (10. *. float changes) ** 2.0 +. float unused ** 2.0 +. cost,
	  min best_bloat (float unused))
    )
    (0.0, 0.0, max_float)
    strings

let dump_status ~gen_ctx ~output (errors, bloat, best_bloat) geno inputs =
  Printf.ksprintf output "Errors: %f Bloat: %f Best bloat: %f Genome: %s\n%!" 
    errors bloat best_bloat
    (String.concat "," (List.map (string_of_operation ~gen_ctx) (List.filter ((<>) Nop) (Array.to_list geno))));
  List.iter
    (fun (s1, s2) ->
       let (str, (changes, unused)) = operate ~gen_ctx s1 geno in
	 Printf.ksprintf output "%s -> %s : %s (%d changes, %d unused)\n"
	   s1 s2
	   str
	   changes
	   unused
    )
    inputs

let generate_function ~fast_stop ~gen_ctx ~output examples =
  let (genomes, fitness) = 
    let stop_iterating = ref None in
      GenePool.evolve 
	{ GenePool.evaluate = evaluate_genome ~gen_ctx examples;
	  compareFitness = 
	    (fun (er1, bl1, bbl1) (er2, bl2, bbl2) ->
	       let c = compare er2 er1 in if c <> 0 then c 
		 else let c = compare bl2 bl1 in if c <> 0 then c else
		   compare bbl2 bbl1
	    );
	  mutate = GenePool.ArrayGenome.mutate 0.2 (mutate_operation ~gen_ctx);
	  crossover = GenePool.ArrayGenome.randomCrossover;
	  genRandom = (fun () -> Array.init 15 (fun _ -> gen_operation ~gen_ctx ()));
	  seed = None;
	  (* report = Some (fun n geno fitness -> dump_status fitness geno examples); *)
	  report = None;
	  stop = 
	    if fast_stop then
	      Some (fun iteration genome (errors, bloat, best_bloat) ->
		      if errors = 0.0 && best_bloat = 0.0 then
			match !stop_iterating with
			  | None -> stop_iterating := Some (iteration / 5 + 30); false
			  | Some iteration' when iteration' >= iteration -> true
			  | _ -> false
		      else
			false
		   )
	    else
	      None
	}
	{ GenePool.nRandom	= 500;
	  nSurvivors		= 100;
	  nMutations		= 300;
	  nCrossovers		= 300;
	  timeLimit		= 10.0;
	  maxGen		= 1000 }
  in
  let geno = genomes.(0) in
  let () = dump_status ~gen_ctx ~output fitness geno examples in
    if let (errors, _, _) = fitness in errors = 0.0 then
      Some geno
    else
      None

let split_words words =
  let patterns = ["[0-9]"; "[ _.-]"] in
    List.unique (
      List.filter ((<>) "") (
	List.concat (
	  List.concat (
	    List.map
	      (fun pat ->
		 List.map (Pcre.split ~pat) words
	      ) patterns
	  ) 
	) @ words
      )
    )
  
let generate_mapping ?(fast_stop=false) ?(output=print_string) examples input =
  let common_words = Words.common_words_list (List.map snd examples) in
  let common_words = split_words common_words in
  let gen_ctx = {
    gc_examples = Array.of_list examples;
    gc_strings = Array.of_list common_words;
  } in
    match generate_function ~fast_stop ~gen_ctx ~output examples with
      | None -> 
	  Printf.printf "Mapping function not found.\n";
	  None
      | Some geno ->
	  let used_names = Hashtbl.create (List.length input) in 
	  let success = ref true in
	  let names =
	    List.map
	      (fun str ->
		 let to_str = fst (operate ~gen_ctx str geno) in
		   if Hashtbl.mem used_names to_str then
		     ( Printf.ksprintf output "ERROR: name %s has already been mapped to %s\n"
			 (Hashtbl.find used_names to_str)
			 to_str;
		       success := false )
		   else
		     Hashtbl.add used_names to_str str;
		   Printf.ksprintf output "%s -> %s\n"
		     str
		     to_str;
		   (str, to_str)
	      )
	      input
	  in
	    if !success then
	      Some names
	    else
	      None
  
